/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLCLM:					*/
/*	Take Care of all "CLAMAV"  anti-virus	*/
/*	scanner access.				*/
/*						*/
/************************************************/
#include	<arpa/inet.h>
#include	<string.h>
#include	<stdio.h>
#include	<time.h>
#include	<setjmp.h>
#include	<errno.h>
#include	<signal.h>
#include	<syslog.h>
#include	<fcntl.h>
#include	<stdlib.h>
#include        <execinfo.h>
#
#include	"lowtyp.h"
#include	"subclm.h"
#include	"unijrl.h"
#include	"unidns.h"
#include	"unieml.h"
#include	"gestls.h"
#include	"lvlcla.h"

#define	CR	"\n"		//EOL = Carriage Retunn

/*how long to wait for Virus scanner		*/
#define	WAITOPEN	10

typedef	struct	{
	time_t last;		/*last load	*/
	char *port;		/*scanner port	*/
	char *hostname;		/*scanner names	*/
	char *version;		/*version used	*/
	char *virname;		/*last virus	*/
	}CLAMFILE;
/*

*/
/************************************************/
/*						*/
/*	Routine to check the result from the	*/
/*	virus scanner				*/
/*						*/
/************************************************/
static CLMCHKTYP lookatscanres(char *result,CLAMFILE *clamfile)

{
#define	OPLO	"lvlcla.c:lookatscanres,"

static VOCTYP vocckr[]={
	{1,"ERROR",(void *)0},
	{2,"FOUND",(void *)0},
	{3,"OK",(void *)0},
	{0,(char *)0,(void *)0}
	};

CLMCHKTYP status;
char *ptr;
int phase;
int proceed;

status=cla_nocheck;
ptr=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*extracting status	*/
      if ((ptr=strrchr(result,' '))==(char *)0) {
        (void) clm_alert(0,"%s no space found within <%s> (bug?)",OPLO,result);
	phase=999;
	}
      break;
    case 1	:	/*extracting status	*/
      if (ptr!=(char *)0) {	/*alway true	*/
	const VOCTYP *voc;

	if ((voc=clm_getvoca(vocckr,ptr+1))!=(const VOCTYP *)0) {
  	  switch (voc->code) {
	    case 1  :	/*Error!		*/
              (void) clm_alert(0,"%s Error found within <%s> (bug?)",
				 OPLO,result);
	      break;
	    case 2  :	/*Virus Found		*/
	      status=cla_virus;
	      break;
	    case 3  :	/*everything OK		*/
	      status=cla_clean;
	      break;
    	    default :	/*big trouble		*/
              (void) clm_alert(0,"%s Unexpected scanner returned result <%s> (bug?)",
				  OPLO,result);
	      break;
	    }
	  }
	}
      break;
    case 2	:	/*extracting virus Name	*/
      clamfile->virname=clm_freestr(clamfile->virname);
      switch (status) {
	case cla_virus	:
	  *ptr='\000';
          if ((ptr=strrchr(result,' '))==(char *)0) {
            (void) clm_alert(0,"%s No virus name found within <%s> (bug?)",OPLO,result);
	    ptr=" Unnamed Virus";
	    }
	  ptr++;
	  clamfile->virname=strdup(ptr);	
	  break;
	case cla_clean	:	/*nothing to do	*/
	case cla_nocheck:
	  break;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to shutdown clamav link		*/
/*						*/
/************************************************/
static TLSFILE *clamshutdown(TLSFILE *tlsfile)

{
#define	CLMSHUTDOWN	"SHUTDOWN"

#define	OPEP	"lvlcla.c:clamshutdown,"

if (tlsfile!=(TLSFILE *)0) {
  if (tls_iseof(tlsfile)==false) 
    (void) tls_printf(tlsfile,"%s%s",CLMSHUTDOWN,CR);
  (void) tls_close(tlsfile);
  }
return (TLSFILE *)0;
#undef OPEP
}
/*
 * 
 * */
/************************************************/
/*						*/
/*	procedure to send a file on specific	*/
/*	clamav scanner port.			*/
/*						*/
/************************************************/
static int streaming(TLSFILE *tls,const char *filename)

{
#define	OPST	"lvlcla.c:streaming,"

#define	CLMINSTREAM	"INSTREAM"/*STREAM file*/

int status;
int handle;
int phase;
int proceed;

status=false;
handle=-1;
phase=0;
proceed=true;
(void) clm_pushdebug(0);//no need to check file sent to clamav
while (proceed==true) {
  switch (phase) {
    case 0	:	//opening file	
      if ((handle=open(filename,O_RDONLY))<0) {
        (void) clm_alert(0,"%s Unable to open file <%s> (error=<%s>)",
			    OPST,filename,strerror(errno));
	proceed=false;
	}
      break;
    case 1	:	/*sending file chuck	*/
      if (handle>=0) {	/*always true		*/
	int stilltosend;

	stilltosend=true;
	status=true;
	(void) tls_printf(tls,"n%s%s",CLMINSTREAM,CR);
	while (stilltosend==true) {
	  size_t got;
	  uint32_t buffer[8192/sizeof(uint32_t)];
	  size_t todo;

	  todo=sizeof(uint32_t);
	  got=read(handle,&buffer[1],sizeof(buffer)-sizeof(uint32_t));
	  if (got>0) {
	    todo+=got;
	    buffer[0]=htonl(got);
	    }
	  else {
	    *buffer=0;
	    stilltosend=false;
	    }
	  if ((got=tls_write(tls,(char *)buffer,todo))!=todo) {
	    status=false;
	    break;	/*not going further	*/
	    }
	  }
        }
      break;
    case 2	:	/*closing file		*/
      if (status==false) {
	char *session;

	if ((session=strrchr(filename,'/'))==(char *)0) 
	   session=((char *)filename)-1;	
	session++;
	(void) clm_alert(0,"%s Unable to fully submit session <%s> to clam daemon (aborting virus detection)",
			   "lvlcal.c:streaming,",session);
	}
      (void) close(handle);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;  
  }
(void) clm_popdebug();
return status;
}
/*

*/
/************************************************/
/*						*/
/*	procedure to send ONE string to a remote*/
/*	TCP server.				*/
/*						*/
/************************************************/
static char *sendoneshot(char *hostname,char *port,int mxwait,const char *fmt,...)

{
#define	OPEP		"lvlcla.c:sendoneshot,"

char *buffer;
va_list args;
TLSFILE *tlsfile;
int socker;
int phase;
int proceed;

buffer=(char *)0;
va_start(args,fmt);
tlsfile=(TLSFILE *)0;
socker=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Opening the channel	*/
      if ((tlsfile=tls_opentcp(hostname,port,WAITOPEN,&socker,CR))==(TLSFILE *)0) {
        (void) clm_alert(3,"%s Unable to access <%s> port <%s> within '%d' sec (config?)",
			    OPEP,hostname,port,WAITOPEN);
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 1	:	/*sending the string	*/
      (void) vasprintf(&buffer,fmt,args);
      (void) tls_printf(tlsfile,"%s%s",buffer,CR);
      (void) free(buffer);
      break;
    case 2	:	/*expecting answer	*/
      if ((buffer=tls_waitforaction(tlsfile,mxwait))==(char *)0) {
	char *local;
	va_list argssup;

	va_start(argssup,fmt);
        (void) vasprintf(&local,fmt,argssup);
        (void) clm_alert(0,"%s No answer to <%s> from scanner within '%d' sec",
			   OPEP,local,mxwait);
	(void) free(local);
	va_end(argssup);
	}
      break;
    case 3	:	/*closing channel	*/
      tlsfile=clamshutdown(tlsfile);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
va_end(args);
return buffer;
#undef OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Routine to get the clamd data-base	*/
/*	Version number				*/
/*						*/
/************************************************/
static int clamdversion(CLAMTYP *clamav)

{
#define	OPEP	"lvlcla.c:clamdversion,"

#define	CLMVERSION	"VERSION"
#define	VERWAIT		30

int status;

status=false;
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *cf;
  char *buffer;
  int phase;
  int proceed;

  cf=(CLAMFILE *)clamav;
  buffer=(char *)0;
  phase=0;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*asking version #	*/
	if ((buffer=sendoneshot(cf->hostname,cf->port,VERWAIT,"%s",CLMVERSION))==(char *)0)
	  phase=999;	/*trouble trouble	*/
	break; 
      case 1	:	/*getting version #	*/
        buffer=clm_cleanstring(buffer);
	cf->version=buffer;
  	cf->last=time((time_t *)0);
	status=true;	
        // fall through
      default	:	/*SAFE Guard		*/
	proceed=false;
	break;
      }
    phase++;
    }
  }
return status;
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a virus scan list	*/
/*						*/
/************************************************/
CLAMTYP *cla_openscanner(char *hostname,char *port)

{
#define	OPEP		"lvlcla.c:cla_openscanner,"

#define	HOSTCAN		"localhost.localdomain"
#define	PORTSCAN	"3310"
#define	CLMSESSION	"SESSION"

CLAMFILE *clamfile;
int phase;
int proceed;

clamfile=(CLAMFILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*setting default value	*/
      if (hostname==(char *)0)
	hostname=HOSTCAN;
      if (port==0)
	port=PORTSCAN;
      clamfile=(CLAMFILE *)calloc(1,sizeof(CLAMFILE));
      clamfile->hostname=strdup(hostname);
      clamfile->port=strdup(port);
      break;
    case 1	:	/*checking reload	*/
      if (clamdversion((CLAMTYP *)clamfile)==false) {
        (void) clm_alert(3,"%s Unable to get clamd data-base version",OPEP);
	clamfile=cla_closescanner(clamfile);
	phase=999;	/*trouble trouble	*/
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return (CLAMTYP *)clamfile;
#undef	OPEP
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free a virus scan list	*/
/*						*/
/************************************************/
CLAMTYP *cla_closescanner(CLAMTYP *clamav)

{
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *clamfile;

  clamfile=(CLAMFILE *)clamav;
  clamfile->virname=clm_freestr(clamfile->virname);
  clamfile->version=clm_freestr(clamfile->version);
  clamfile->port=clm_freestr(clamfile->port);
  clamfile->hostname=clm_freestr(clamfile->hostname);
  (void) free(clamfile);
  clamfile=(CLAMFILE *)0;
  clamav=(CLAMTYP *)clamfile;
  }
return clamav;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if a virus is found	*/
/*	within the E-mail contents.		*/
/*	If virus discovered, return the virus	*/
/*	name.					*/
/*						*/
/************************************************/
CLMCHKTYP cla_checkvirus(CLAMTYP **clamav,char *filename,const char **virname)

{
#define	OPCH		"lvlcla.c:cla_checkvirus,"
#define	CLMWSCAN	30	/*wait 30 sec 	*/
#define	CLMIDSESSION	"IDSESSION"
#define	CLMEND		"END"

CLMCHKTYP status;

status=cla_nocheck;
*virname=(char *)0;
if (*clamav!=(CLAMTYP *)0) {
  CLAMFILE *cf;
  TLSFILE *tlsfile;
  char *buffer;
  int socker;
  int phase;
  int proceed;

  cf=(CLAMFILE *)*clamav;
  tlsfile=(TLSFILE *)0;
  buffer=(char *)0;
  socker=0;
  phase=1;
  proceed=true;
  while (proceed==true) {
    switch (phase) {
      case 0	:	/*we have a big trouble	*/
        (void) clm_alert(0,"%s Lost contact with clamd daemon <%s/%d>!",
			    OPCH,cf->hostname,cf->port);
	*clamav=cla_closescanner(*clamav);
	proceed=false;
        break;
      case 1	:	/*opening channel	*/
        if ((tlsfile=tls_opentcp(cf->hostname,cf->port,WAITOPEN,&socker,CR))==(TLSFILE *)0) {
          (void) clm_alert(0,"%s Unable to access <%s> port <%s> within '%d' sec (config?)",
			      OPCH,cf->hostname,cf->port,WAITOPEN);
	  phase=-1;	/*trouble trouble	*/
          }
        break;
      case 2	:	/*waiting answer	*/
        (void) tls_printf(tlsfile,"n%s%s",CLMIDSESSION,CR);
	break;
      case 3	:	/*streaming file 	*/
	if (streaming(tlsfile,filename)==false) {
          (void) clm_alert(0,"%s Aborting virus detection",OPCH);
	  tlsfile=clamshutdown(tlsfile);
	  phase=-1;	/*trouble trouble	*/
	  }
	break;
      case 4	:	/*expecting answer	*/
        if ((buffer=tls_waitforaction(tlsfile,CLMWSCAN))==(char *)0) {
          (void) clm_alert(0,"%s No answer scanning file <%s> within '%d' sec",
			      OPCH,filename,CLMWSCAN);
	  tlsfile=clamshutdown(tlsfile);
          phase=-1;	/*trouble trouble	*/
	  }
	break;
      case 5	:	/*checking the answer	*/
	buffer=clm_cleanstring(buffer);
	switch (status=lookatscanres(buffer,cf)) {
	  case cla_clean	:
	    break;	/*nothing to do		*/
	  case cla_virus	:
	    *virname=cf->virname;
	    break;	/*get virusname		*/
	  case cla_nocheck	:
	    break;	/*nothing to do		*/
	  }
        (void) free(buffer);
	break;
      default	:	/*SAFE Guard		*/
        tlsfile=clamshutdown(tlsfile);
	proceed=false;
	break;
      }
    phase++;
    }
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to return the clamd version	*/
/*	and the last update date.		*/
/*						*/
/************************************************/
char *cla_getversdate(CLAMTYP *clamav)

{
char *versdate;

versdate=(char *)0;
if (clamav!=(CLAMTYP *)0) {
  CLAMFILE *clamfile;

  clamfile=(CLAMFILE *)clamav;
  (void) asprintf(&versdate,"%d %s",CMDOK,clamfile->version);
  }
return versdate;
}
/*

*/
/************************************************/
/*						*/
/*	Routine to check if a new virus tree is	*/
/*	needed to be loaded			*/
/*						*/
/************************************************/
int cla_dbreload(char *hostname,char *port)

{
#define	OPEP	"lvlcla.c:cla_dbreload,"

#define	CLMRELOAD	"RELOAD"
#define	CLMRELOADING	"RELOADING"
#define	WRELOAD		60

int status;
CLAMTYP *clamav;
int phase;
int proceed;

status=false;
clamav=(CLAMTYP *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Opening Virus scanner	*/
      if ((clamav=cla_openscanner(hostname,port))==(CLAMTYP *)0) {
        (void) clm_alert(0,"%s Unable to get scanner data host <%s> port <%s>",
			    OPEP,hostname,port);
	phase=999;	/*no need to further	*/
	}
      break;
    case 1	:	/*reloading		*/
      status=true;
      break;
    default	:	/*SAFE Guard		*/
      if (clamav!=(CLAMTYP *)0) 
	clamav=cla_closescanner(clamav);
      proceed=false;
      break;
    }
  phase++;
  }
return status;
#undef	OPEP
}
